@@ -28,7 +28,7 @@
 require_once(__DIR__ . '/../../behat/behat_base.php');
 
 use Behat\Gherkin\Node\TableNode as TableNode;
-use Behat\Behat\Exception\PendingException as PendingException;
+use Behat\Behat\Tester\Exception\PendingException as PendingException;
 
 /**
  * Class to set up quickly a Given environment.
@@ -110,7 +110,11 @@ class behat_data_generators extends behat_base {
         'activities' => array(
             'datagenerator' => 'activity',
             'required' => array('activity', 'idnumber', 'course'),
-            'switchids' => array('course' => 'course', 'gradecategory' => 'gradecat')
+            'switchids' => array('course' => 'course', 'gradecategory' => 'gradecat', 'grouping' => 'groupingid')
+        ),
+        'blocks' => array(
+            'datagenerator' => 'block_instance',
+            'required' => array('blockname', 'contextlevel', 'reference'),
         ),
         'group members' => array(
             'datagenerator' => 'group_member',
@@ -159,7 +163,7 @@ class behat_data_generators extends behat_base {
         'question categories' => array(
             'datagenerator' => 'question_category',
             'required' => array('name', 'contextlevel', 'reference'),
-            'switchids' => array('questioncategory' => 'category')
+            'switchids' => array('questioncategory' => 'parent')
         ),
         'questions' => array(
             'datagenerator' => 'question',
@@ -170,10 +174,22 @@ class behat_data_generators extends behat_base {
             'datagenerator' => 'tag',
             'required' => array('name')
         ),
+        'events' => array(
+            'datagenerator' => 'event',
+            'required' => array('name', 'eventtype'),
+            'switchids' => array(
+                'user' => 'userid',
+                'course' => 'courseid',
+                'category' => 'categoryid',
+            )
+        ),
     );
 
     /**
-     * Creates the specified element. More info about available elements in http://docs.moodle.org/dev/Acceptance_testing#Fixtures.
+     * Creates the specified element.
+     *
+     * The most reliable list of what types of thing can be created is the
+     * $elements array defined above.
      *
      * @Given /^the following "(?P<element_string>(?:[^"]|\\")*)" exist:$/
      *
@@ -341,6 +357,39 @@ class behat_data_generators extends behat_base {
         }
     }
 
+    /**
+     * Add a block to a page.
+     *
+     * @param array $data should mostly match the fields of the block_instances table.
+     *     The block type is specified by blockname.
+     *     The parentcontextid is set from contextlevel and reference.
+     *     Missing values are filled in by testing_block_generator::prepare_record.
+     *     $data is passed to create_block as both $record and $options. Normally
+     *     the keys are different, so this is a way to let people set values in either place.
+     */
+    protected function process_block_instance($data) {
+
+        if (empty($data['blockname'])) {
+            throw new Exception('\'blocks\' requires the field \'block\' type to be specified');
+        }
+
+        if (empty($data['contextlevel'])) {
+            throw new Exception('\'blocks\' requires the field \'contextlevel\' to be specified');
+        }
+
+        if (!isset($data['reference'])) {
+            throw new Exception('\'blocks\' requires the field \'reference\' to be specified');
+        }
+
+        $context = $this->get_context($data['contextlevel'], $data['reference']);
+        $data['parentcontextid'] = $context->id;
+
+        // Pass $data as both $record and $options. I think that is unlikely to
+        // cause problems since the relevant key names are different.
+        // $options is not used in most blocks I have seen, but where it is, it is necessary.
+        $this->datagenerator->create_block($data['blockname'], $data, $data);
+    }
+
     /**
      * Adapter to enrol_user() data generator.
      * @throws Exception
@@ -515,7 +564,31 @@ class behat_data_generators extends behat_base {
      * @param array $data the row of data from the behat script.
      */
     protected function process_question_category($data) {
+        global $DB;
+
         $context = $this->get_context($data['contextlevel'], $data['reference']);
+
+        // The way this class works, we have already looked up the given parent category
+        // name and found a matching category. However, it is possible, particularly
+        // for the 'top' category, for there to be several categories with the
+        // same name. So far one will have been picked at random, but we need
+        // the one from the right context. So, if we have the wrong category, try again.
+        // (Just fixing it here, rather than getting it right first time, is a bit
+        // of a bodge, but in general this class assumes that names are unique,
+        // and normally they are, so this was the easiest fix.)
+        if (!empty($data['parent'])) {
+            $foundparent = $DB->get_record('question_categories', ['id' => $data['parent']], '*', MUST_EXIST);
+            if ($foundparent->contextid != $context->id) {
+                $rightparentid = $DB->get_field('question_categories', 'id',
+                        ['contextid' => $context->id, 'name' => $foundparent->name]);
+                if (!$rightparentid) {
+                    throw new Exception('The specified question category with name "' . $foundparent->name .
+                            '" does not exist in context "' . $context->get_context_name() . '"."');
+                }
+                $data['parent'] = $rightparentid;
+            }
+        }
+
         $data['contextid'] = $context->id;
         $this->datagenerator->get_plugin_generator('core_question')->create_question_category($data);
     }
@@ -658,6 +731,11 @@ class behat_data_generators extends behat_base {
     protected function get_grouping_id($idnumber) {
         global $DB;
 
+        // Do not fetch grouping ID for empty grouping idnumber.
+        if (empty($idnumber)) {
+            return null;
+        }
+
         if (!$id = $DB->get_field('groupings', 'id', array('idnumber' => $idnumber))) {
             throw new Exception('The specified grouping with idnumber "' . $idnumber . '" does not exist');
         }
